#pragma once

#include "volumeshop.h"

#include <list>
#include <set>

#include "RendererPlugin.h"
#include "PluginInstance.h"
#include "Volume.h"
#include "Renderer.h"
#include "Vector.h"
#include "Quaternion.h"
#include "Matrix.h"
#include "Shader.h"
#include "Color.h"
#include "Timer.h"
#include "Label.h"
#include "Exception.h"
#include "Primitive.h"
#include "Plugin.h"
#include "Image.h"
#include "Volume.h"
#include "Voxel.h"

class RendererVolumeLao : public Renderer, public PluginInstance
{
	class BasicGenerator
	{
	public:
		virtual void generate(std::vector<float> & vecVertices, std::vector<unsigned int> & vecIndices, const Image<unsigned char,4> & imgImage) = 0;
	};

	template < template<class,unsigned int> class ResourceType, class ElementType, unsigned int N>
	class Generator : public BasicGenerator
	{
	public:
		Generator(Resource & resResource)
		{
		};

		~Generator()
		{
		};

		virtual void generate(std::vector<float> & vecVertices, std::vector<unsigned int> & vecIndices, const Image<unsigned char,4> & imgImage)
		{
		};
	};

	template <class ElementType, unsigned int N>
	class Generator<Volume,ElementType,N> : public BasicGenerator
	{
	public:
		Generator(Resource & resResource) : m_volVolume(resResource.GetType< Volume<ElementType,N> >())
		{
		};

		~Generator()
		{
		};

		virtual void generate(std::vector<float> & vecVertices, std::vector<unsigned int> & vecIndices, const Image<unsigned char,4> & imgImage)
		{
			Volume<ElementType,N>::HierachicalIterator iter(m_volVolume);
			Image<unsigned char,4>::IntegralIterator intIntegral(imgImage);
			generate(vecVertices,vecIndices,imgImage,iter);
		};

	private:

		void generate(std::vector<float> & vecVertices, std::vector<unsigned int> & vecIndices, const Image<unsigned char,4>::IntegralIterator & intIntegral, typename const Volume<ElementType,N>::HierachicalIterator & iter)
		{
			static const unsigned int vuIndices[6][4] =
			{

				{0, 1, 2, 3},
				{3, 2, 6, 7},
				{7, 6, 5, 4},
				{4, 5, 1, 0},
				{5, 6, 2, 1},
				{7, 4, 0, 3}
			};

			const unsigned int uVolumeWidth = m_volVolume.GetWidth();
			const unsigned int uVolumeHeight = m_volVolume.GetHeight();
			const unsigned int uVolumeDepth = m_volVolume.GetDepth();

			const unsigned int uX = iter.GetPositionX();
			const unsigned int uY = iter.GetPositionY();
			const unsigned int uZ = iter.GetPositionZ();

			const unsigned int uWidth = iter.GetWidth();
			const unsigned int uHeight = iter.GetHeight();
			const unsigned int uDepth = iter.GetDepth();

			const unsigned int uStartX = uX - uWidth/2;
			const unsigned int uStartY = uY - uHeight/2;
			const unsigned int uStartZ = uZ - uDepth/2;

			if (uStartX > m_volVolume.GetWidth() || uStartY > m_volVolume.GetHeight() || uStartZ > m_volVolume.GetDepth())
				return;

			const unsigned int uEndX = std::min(uVolumeWidth,uX + uWidth/2);
			const unsigned int uEndY = std::min(uVolumeHeight,uY + uHeight/2);
			const unsigned int uEndZ = std::min(uVolumeDepth,uZ + uDepth/2);

			float vfVertices[8][3];
			vfVertices[0][0] = vfVertices[1][0] = vfVertices[2][0] = vfVertices[3][0] = float(uStartX);
			vfVertices[4][0] = vfVertices[5][0] = vfVertices[6][0] = vfVertices[7][0] = float(uEndX) - 1.0f/16384.0f;
			vfVertices[0][1] = vfVertices[1][1] = vfVertices[4][1] = vfVertices[5][1] = float(uStartY);
			vfVertices[2][1] = vfVertices[3][1] = vfVertices[6][1] = vfVertices[7][1] = float(uEndY) - 1.0f/16384.0f;
			vfVertices[0][2] = vfVertices[3][2] = vfVertices[4][2] = vfVertices[7][2] = float(uStartZ);
			vfVertices[1][2] = vfVertices[2][2] = vfVertices[5][2] = vfVertices[6][2] = float(uEndZ) - 1.0f/16384.0f;


			const unsigned int uOffset = vecVertices.size() / 3;

			for (int i=0;i<24;i++)
			{
				vecVertices.push_back(*(&(vfVertices[0][0])+i));
				vecIndices.push_back(*(&(vuIndices[0][0])+i)+uOffset);
			}
		};

	private:
		Volume<ElementType, N> & m_volVolume;
	};

	template <class ElementType>
	class Generator<Volume,ElementType,1> : public BasicGenerator
	{
	public:
		Generator(Resource & resResource) : m_volVolume(resResource.GetType< Volume<ElementType,1> >())
		{
		};

		~Generator()
		{
		};

		virtual void generate(std::vector<float> & vecVertices, std::vector<unsigned int> & vecIndices, const Image<unsigned char,4> & imgImage)
		{
			Volume<ElementType,1>::HierachicalIterator iter(m_volVolume);
			Image<unsigned char,4>::IntegralIterator intIntegral(imgImage);
			generate(vecVertices,vecIndices,imgImage,iter);
		};
	
	private:

		void generate(std::vector<float> & vecVertices, std::vector<unsigned int> & vecIndices, const Image<unsigned char,4>::IntegralIterator & intIntegral, typename const Volume<ElementType,1>::HierachicalIterator & iter)
		{
			const Voxel<float,1> voxMinimum = iter.GetMinimum();
			const Voxel<float,1> voxMaximum = iter.GetMaximum();

			const unsigned int u0x = unsigned int(voxMinimum.Get(0)*float(intIntegral.GetWidth()-1));
			const unsigned int u1x = unsigned int(voxMaximum.Get(0)*float(intIntegral.GetWidth()-1));
			const Voxel<float,4> voxIntegral = intIntegral.GetIntegral(u0x,u1x);

			if (voxIntegral.Get(3) > 1.192092896e-07F)
			{
				if (iter.IsLeaf())
				{
					static const unsigned int vuIndices[6][4] =
					{

						{0, 1, 2, 3},
						{3, 2, 6, 7},
						{7, 6, 5, 4},
						{4, 5, 1, 0},
						{5, 6, 2, 1},
						{7, 4, 0, 3}
					};

					const unsigned int uVolumeWidth = m_volVolume.GetWidth();
					const unsigned int uVolumeHeight = m_volVolume.GetHeight();
					const unsigned int uVolumeDepth = m_volVolume.GetDepth();

					const unsigned int uX = iter.GetPositionX();
					const unsigned int uY = iter.GetPositionY();
					const unsigned int uZ = iter.GetPositionZ();

					const unsigned int uWidth = iter.GetWidth();
					const unsigned int uHeight = iter.GetHeight();
					const unsigned int uDepth = iter.GetDepth();

					const unsigned int uStartX = uX - uWidth/2;
					const unsigned int uStartY = uY - uHeight/2;
					const unsigned int uStartZ = uZ - uDepth/2;
				
					if (uStartX > m_volVolume.GetWidth() || uStartY > m_volVolume.GetHeight() || uStartZ > m_volVolume.GetDepth())
						return;

					const unsigned int uEndX = std::min(uVolumeWidth,uX + uWidth/2);
					const unsigned int uEndY = std::min(uVolumeHeight,uY + uHeight/2);
					const unsigned int uEndZ = std::min(uVolumeDepth,uZ + uDepth/2);

					float vfVertices[8][3];
					vfVertices[0][0] = vfVertices[1][0] = vfVertices[2][0] = vfVertices[3][0] = float(uStartX);
					vfVertices[4][0] = vfVertices[5][0] = vfVertices[6][0] = vfVertices[7][0] = float(uEndX) - 1.0f/16384.0f;
					vfVertices[0][1] = vfVertices[1][1] = vfVertices[4][1] = vfVertices[5][1] = float(uStartY);
					vfVertices[2][1] = vfVertices[3][1] = vfVertices[6][1] = vfVertices[7][1] = float(uEndY) - 1.0f/16384.0f;
					vfVertices[0][2] = vfVertices[3][2] = vfVertices[4][2] = vfVertices[7][2] = float(uStartZ);
					vfVertices[1][2] = vfVertices[2][2] = vfVertices[5][2] = vfVertices[6][2] = float(uEndZ) - 1.0f/16384.0f;


					const unsigned int uOffset = vecVertices.size() / 3;

					for (int i=0;i<24;i++)
					{
						vecVertices.push_back(*(&(vfVertices[0][0])+i));
						vecIndices.push_back(*(&(vuIndices[0][0])+i)+uOffset);
					}
				}
				else
				{
					for (Volume<ElementType,1>::HierachicalIterator i(iter); !i.IsAtEnd(); ++i)
						generate(vecVertices,vecIndices,intIntegral,(*i));
				}
			}
		};

	private:
		Volume<ElementType, 1> & m_volVolume;
	};


public:

	RendererVolumeLao(Plugin & pluPlugin) : PluginInstance(pluPlugin), m_uWidth(1), m_uHeight(1), 
		m_uVolumeFrameBuffer(0), m_uVolumeColorTexture(0), m_uGeometryDepthTexture(0), m_uVolumeDepthRenderBuffer(0), m_uFrontColorTexture(0), m_uBackColorTexture(0),
		m_uVertexBuffer(0), m_uIndexBuffer(0), m_bGenerateBounds(true), 
		m_shaEntry(pluPlugin.GetPluginDirectory()+"plugin_renderer_volumelao_entry.glsl"),
		m_shaGeometry(pluPlugin.GetPluginDirectory()+"plugin_renderer_volumelao_geometry.glsl"),
		m_shaVolume(pluPlugin.GetPluginDirectory()+"plugin_renderer_volumelao_volume.glsl")
	{
//		z_scale=2.0;
		if (!GLEW_ARB_fragment_shader)
			throw Exception("GL_ARB_fragment_shader is not supported by your hardware or driver.");

		if (!GLEW_ARB_vertex_shader)
			throw Exception("GL_ARB_fragment_shader is not supported by your hardware or driver.");

		if (!GLEW_EXT_framebuffer_object)
			throw Exception("GL_EXT_framebuffer_object is not supported by your hardware or driver.");

		m_modVariantObserverGenerate.connect(this,&RendererVolumeLao::propertyModifiedGenerate);
		m_modVariantObserverUpdate.connect(this,&RendererVolumeLao::propertyModifiedUpdate);

		glGenFramebuffersEXT(1, &m_uVolumeFrameBuffer);
		glGenTextures(1,&m_uVolumeColorTexture);
		glGenTextures(1,&m_uFrontColorTexture);
		glGenTextures(1,&m_uBackColorTexture);
		glGenTextures(1,&m_uGeometryDepthTexture);
		glGenRenderbuffersEXT(1, &m_uVolumeDepthRenderBuffer); 	

		glGenBuffers( 1, &m_uVertexBuffer );
		glGenBuffers( 1, &m_uIndexBuffer );

		srand(7);
		const unsigned int uDimension = 32;
		unsigned char vuNoise[uDimension*uDimension];
/*
		//########

		char ambientTexture[128*128*128];
		glGenTextures(1,&ambientTexture);

		//########

*/
		printf("test-end\n");
		printf("Size of float %d\n", sizeof(float) );
/*
		GLfloat emptyTexture[128][128][128];
		for (int i= 0; i<128;i++){
			for (int j= 0; j<128;j++){
				for (int k= 0; k<128;k++){
					emptyTexture[i][j][k]=(GLfloat)0.0;
				}
			}
		}
*/
/*
		GLfloat emptyTexture[128*128*128];
		for (int i= 0; i<128*128*128;i++){
			emptyTexture[i]=(GLfloat)0.0;
		}
*/
		glEnable(GL_TEXTURE_3D);
		glGenTextures(1,&m_uAmbientTexture);
		glBindTexture(GL_TEXTURE_3D,m_uAmbientTexture);

		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

		//gl.glTexImage3D(GL_TEXTURE_3D, 0, GL_RGBA32F_ARB, width, height, depth,  0, GL_RGBA, GL_FLOAT, null);
		//glTexImage3D(GL_TEXTURE_3D, 0, GL_RGBA32F_ARB, 128, 128, 128,  0, GL_RGBA, GL_FLOAT, NULL);
		glTexImage3D(GL_TEXTURE_3D, 0, GL_ALPHA, 128, 128, 128,  0, GL_ALPHA, GL_FLOAT, NULL);
		//glFramebufferTexture3DEXT(GL_FRAMEBUFFER_EXT,  GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_3D, m_uAmbientTexture, 0, 0); 
		glBindTexture(GL_TEXTURE_3D,0);


		printf("test-end\n");



		for (int i=0;i<uDimension*uDimension;i++)
			vuNoise[i] = rand() / 128;

		glGenTextures(1,&m_uNoiseTexture);
		glBindTexture(GL_TEXTURE_2D,m_uNoiseTexture);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
		glTexImage2D(GL_TEXTURE_2D,0,GL_ALPHA,uDimension,uDimension,0,GL_ALPHA,GL_UNSIGNED_BYTE,vuNoise);
		glBindTexture(GL_TEXTURE_2D,0);

		GetPlugin().GetProperty("Projection Transformation").require(Variant::TypeMatrix());
		GetPlugin().GetProperty("Projection Transformation").addObserver(&m_modVariantObserverUpdate);

		GetPlugin().GetProperty("Viewing Transformation").require(Variant::TypeMatrix());
		GetPlugin().GetProperty("Viewing Transformation").addObserver(&m_modVariantObserverUpdate);

		GetPlugin().GetProperty("Volume").require(Variant::TypeHandle());
		GetPlugin().GetProperty("Volume").addObserver(&m_modVariantObserverGenerate);

		GetPlugin().GetProperty("Transfer Function").require(Variant::TypeHandle());
		GetPlugin().GetProperty("Transfer Function").addObserver(&m_modVariantObserverGenerate);

		GetPlugin().GetProperty("Technique").require((Variant::TypeOption(), Variant("DVR"), Variant("MIP"), Variant("Ambient") ));
		GetPlugin().GetProperty("Technique").addObserver(&m_modVariantObserverUpdate);

		GetPlugin().GetProperty("Selectable").require((Variant::TypeOption(), Variant("NoSelect14"), Variant("NoSelect26"), Variant("Select")));
		GetPlugin().GetProperty("Selectable").addObserver(&m_modVariantObserverUpdate);


		GetPlugin().GetProperty("Sample Distance").require(Variant::TypeFloat(0.25f,0.25f,2.0f));
		GetPlugin().GetProperty("Sample Distance").addObserver(&m_modVariantObserverUpdate);

		GetPlugin().GetProperty("N steps").require(Variant::TypeInteger(20,1,40));
		GetPlugin().GetProperty("N steps").addObserver(&m_modVariantObserverUpdate);

		GetPlugin().GetProperty("N rays").require(Variant::TypeInteger(14,1,26));
		GetPlugin().GetProperty("N rays").addObserver(&m_modVariantObserverUpdate);

		GetPlugin().GetProperty("Steplength").require(Variant::TypeInteger(1,1,3));
		GetPlugin().GetProperty("Steplength").addObserver(&m_modVariantObserverUpdate);


		GetPlugin().GetProperty("Z scale").require(Variant::TypeFloat(1.0f,0.0f,3.0f));
		GetPlugin().GetProperty("Z scale").addObserver(&m_modVariantObserverUpdate);


		GetPlugin().GetProperty("Opacity").require(Variant::TypeFloat(1.0f,0.0f,1.0f));
		GetPlugin().GetProperty("Opacity").addObserver(&m_modVariantObserverUpdate);

		GetPlugin().GetProperty("Z cut").require(Variant::TypeFloat(0.0f,0.0f,1.0f));
		GetPlugin().GetProperty("Z cut").addObserver(&m_modVariantObserverUpdate);

		GetPlugin().GetProperty("Y cut").require(Variant::TypeFloat(0.0f,0.0f,1.0f));
		GetPlugin().GetProperty("Y cut").addObserver(&m_modVariantObserverUpdate);
		
		GetPlugin().GetProperty("X cut").require(Variant::TypeFloat(0.0f,0.0f,1.0f));
		GetPlugin().GetProperty("X cut").addObserver(&m_modVariantObserverUpdate);


	};
 
	virtual ~RendererVolumeLao()
	{
		if (m_uIndexBuffer)
			glDeleteBuffers(1, &m_uIndexBuffer );

		if (m_uVertexBuffer)
			glDeleteBuffers(1, &m_uVertexBuffer );

		if (m_uFrontColorTexture)
			glDeleteTextures(1,&m_uFrontColorTexture);

		if (m_uBackColorTexture)
			glDeleteTextures(1,&m_uBackColorTexture);

		if (m_uVolumeColorTexture)
			glDeleteTextures(1,&m_uVolumeColorTexture);

		if (m_uGeometryDepthTexture)
			glDeleteTextures(1,&m_uGeometryDepthTexture);

		if (m_uVolumeDepthRenderBuffer)
			glDeleteRenderbuffersEXT(1, &m_uVolumeDepthRenderBuffer);

		if (m_uVolumeFrameBuffer)
			glDeleteFramebuffersEXT(1, &m_uVolumeFrameBuffer);
	};

	virtual void idle()
	{
	};

	virtual void display(Canvas & canCanvas)
	{
		renderVolume(canCanvas);
	};

	virtual void reshape(const unsigned int uWidth, const unsigned int uHeight)
	{
		GLenum glStatus;

		m_uWidth = uWidth;
		m_uHeight = uHeight;

		//////////////////////////////////////////////////////////////////////////
		glBindTexture(GL_TEXTURE_2D,m_uVolumeColorTexture);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA16F_ARB,uWidth,uHeight,0,GL_RGBA,GL_FLOAT,NULL);
		glBindTexture(GL_TEXTURE_2D,0);
		//////////////////////////////////////////////////////////////////////////
		glBindTexture(GL_TEXTURE_2D,m_uFrontColorTexture);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
		glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA16F_ARB,uWidth,uHeight,0,GL_RGBA,GL_FLOAT,NULL);
		glBindTexture(GL_TEXTURE_2D,0);
		//////////////////////////////////////////////////////////////////////////
		glBindTexture(GL_TEXTURE_2D,m_uBackColorTexture);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
		glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA16F_ARB,uWidth,uHeight,0,GL_RGBA,GL_FLOAT,NULL);
		glBindTexture(GL_TEXTURE_2D,0);
		//////////////////////////////////////////////////////////////////////////	
		glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, m_uVolumeDepthRenderBuffer);
		glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT,GL_DEPTH_COMPONENT, uWidth, uHeight);  
		glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0);
		//////////////////////////////////////////////////////////////////////////
		glBindTexture(GL_TEXTURE_2D,m_uGeometryDepthTexture);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
		glTexImage2D(GL_TEXTURE_2D,0,GL_DEPTH_COMPONENT,uWidth,uHeight,0,GL_DEPTH_COMPONENT,GL_FLOAT,NULL);
		glBindTexture(GL_TEXTURE_2D,0);
		//////////////////////////////////////////////////////////////////////////
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_uVolumeFrameBuffer);
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT,GL_TEXTURE_2D, m_uVolumeColorTexture, 0);
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT1_EXT,GL_TEXTURE_2D, m_uFrontColorTexture, 0);
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT2_EXT,GL_TEXTURE_2D, m_uBackColorTexture, 0);
		glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT,GL_RENDERBUFFER_EXT, m_uVolumeDepthRenderBuffer);

		glStatus = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
		if (glStatus  != GL_FRAMEBUFFER_COMPLETE_EXT)
			throw Exception("Error initializing volume framebuffer.");

		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
		//////////////////////////////////////////////////////////////////////////

		GetPlugin().update();
	};


	virtual void button(const Controller & conController)
	{
	};

	virtual void cursor(const Controller & conController)
	{
	};

protected:

	virtual void renderVolume(Canvas & canCanvas)
	{
/*
		ray_steps = GetPlugin().GetProperty("N steps").require(Variant::TypeInteger(20,1,40));
		n_rays = GetPlugin().GetProperty("N rays").require(Variant::TypeInteger(14,1,26));
		steplength = GetPlugin().GetProperty("Steplength").require(Variant::TypeInteger(1,1,3));
*/
		ray_steps = GetPlugin().GetProperty("N steps").require(Variant::TypeInteger(20,1,40));
		n_rays = GetPlugin().GetProperty("N rays").require(Variant::TypeInteger(14,1,26));
		steplength = GetPlugin().GetProperty("Steplength").require(Variant::TypeInteger(1,1,3));




		z_scale = GetPlugin().GetProperty("Z scale").require(Variant::TypeFloat(1.0f,0.0f,3.0f));
		z_cut = GetPlugin().GetProperty("Z cut").require(Variant::TypeFloat(0.0f,0.0f,1.0f));
		y_cut = GetPlugin().GetProperty("Y cut").require(Variant::TypeFloat(0.0f,0.0f,1.0f));
		x_cut = GetPlugin().GetProperty("X cut").require(Variant::TypeFloat(0.0f,0.0f,1.0f));
		const Handle hanVolume = GetPlugin().GetProperty("Volume").require(Variant::TypeHandle());

		if (!hanVolume.IsValid())
			return;

		const Handle hanImage = GetPlugin().GetProperty("Transfer Function").require(Variant::TypeHandle());

		if (!hanImage.IsValid() && hanVolume->GetElementComponents() == 1)
			return;

		const Matrix matVolumeTransformation = hanVolume->GetProperty("Transformation").require(Variant::TypeMatrix());
		const Matrix matProjectionTransformation = GetPlugin().GetProperty("Projection Transformation").require(Variant::TypeMatrix());
		const Matrix matViewingTransformation = GetPlugin().GetProperty("Viewing Transformation").require(Variant::TypeMatrix());
		const float fSampleDistance = GetPlugin().GetProperty("Sample Distance").require(Variant::TypeFloat(0.25f,0.25f,2.0f));

		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D,m_uGeometryDepthTexture);
		glCopyTexSubImage2D(GL_TEXTURE_2D,0,0,0,0,0,m_uWidth,m_uHeight);
		glBindTexture(GL_TEXTURE_2D,0);

		glActiveTexture(GL_TEXTURE2);
		hanVolume->bind();

		glActiveTexture(GL_TEXTURE3);
		if (hanImage.IsValid()) hanImage->bind();

		glActiveTexture(GL_TEXTURE4);
		glBindTexture(GL_TEXTURE_2D,m_uNoiseTexture);

		glActiveTexture(GL_TEXTURE5);
		glBindTexture(GL_TEXTURE_3D,m_uAmbientTexture);

	
		Matrix matDataTransformation;
		const Vector vecMaximum = hanVolume->GetBounds().GetMaximum();
		const float fDimension = hanVolume->GetSize().GetMaximum();
		matDataTransformation *= matVolumeTransformation;
		matDataTransformation.scale(Vector(1.0f / fDimension,1.0f / fDimension,z_scale*1.0f / fDimension));
		matDataTransformation.translate(vecMaximum * -0.5f);

		Matrix matTextureTransformation;
		matTextureTransformation = hanVolume->GetTextureTransformation() * (matViewingTransformation*matDataTransformation).GetInverse();

		Matrix matGeometryTransformation;
		matGeometryTransformation = (matTextureTransformation * matViewingTransformation) * (matProjectionTransformation * matViewingTransformation).GetInverse();

		Matrix matModelViewTransformation = matViewingTransformation*matDataTransformation;
		Matrix matInverseModelViewTransformation = matModelViewTransformation.GetInverse();

		Matrix matModelViewProjectionTransformation = matProjectionTransformation * matModelViewTransformation;
		Matrix matInverseModelViewProjectionTransformation = matModelViewProjectionTransformation.GetInverse();

		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_uVolumeFrameBuffer);

		glMatrixMode(GL_PROJECTION);
		glLoadMatrixf(matProjectionTransformation.Get());

		glMatrixMode(GL_MODELVIEW);
		glLoadMatrixf(matViewingTransformation.Get());
		glMultMatrixf(matDataTransformation.Get());

		m_shaEntry.bind();
		glUniformMatrix4fv(m_shaEntry.GetUniformLocation("matTexture"),1,false,matTextureTransformation.Get());

		//////////////////////////////////////////////////////////////////////////
		// Render front faces
		//////////////////////////////////////////////////////////////////////////
		glDrawBuffer(GL_COLOR_ATTACHMENT1_EXT);

		glClearColor(0.0f,0.0f,0.0f,1.0f);
		glClearDepth(1.0f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		glEnable(GL_CULL_FACE);
		glEnable(GL_DEPTH_TEST);

		glDepthFunc(GL_LESS);
		glCullFace(GL_BACK);

		renderBounds();
		//////////////////////////////////////////////////////////////////////////

		//////////////////////////////////////////////////////////////////////////
		// Render back faces
		//////////////////////////////////////////////////////////////////////////
		glDrawBuffer(GL_COLOR_ATTACHMENT2_EXT);

		glClearColor(0.0f,0.0f,0.0f,0.0f);
		glClearDepth(0.0);
		glClear(GL_COLOR_BUFFER_BIT);

		glEnable(GL_DEPTH_TEST);
		glEnable(GL_CULL_FACE);

		glDepthFunc(GL_GREATER);
		glCullFace(GL_FRONT);		

		renderBounds();
		//////////////////////////////////////////////////////////////////////////

		m_shaEntry.release();

		m_shaGeometry.bind();
		glUniform1iARB(m_shaGeometry.GetUniformLocation("samImage"), 0);
		glUniformMatrix4fv(m_shaGeometry.GetUniformLocation("matGeometry"),1,false,matGeometryTransformation.Get());

		//////////////////////////////////////////////////////////////////////////
		// Render geometry
		//////////////////////////////////////////////////////////////////////////
		glDisable(GL_CULL_FACE);
		glDrawBuffer(GL_COLOR_ATTACHMENT2_EXT);

		glEnable(GL_DEPTH_TEST);
		glDepthFunc(GL_LESS);

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D,m_uGeometryDepthTexture);

		glBegin(GL_QUADS);
		glTexCoord2f(0.0f,0.0f);
		glVertex3f(-1.0f,-1.0f,0.0f);
		glTexCoord2f(1.0f,0.0f);
		glVertex3f(1.0f,-1.0f,0.0f);
		glTexCoord2f(1.0f,1.0f);
		glVertex3f(1.0f,1.0f,0.0f);
		glTexCoord2f(0.0f,1.0f);
		glVertex3f(-1.0f,1.0f,0.0f);
		glEnd();

		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D,0);

		m_shaGeometry.release();
		//////////////////////////////////////////////////////////////////////////

		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);

		canCanvas.bind();

		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

		glEnable(GL_DEPTH_TEST);
		glDepthFunc(GL_LESS);

		glActiveTexture(GL_TEXTURE0);		
		glBindTexture(GL_TEXTURE_2D,m_uFrontColorTexture);

		glActiveTexture(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D,m_uBackColorTexture);

		const Vector vecView = matViewingTransformation.GetInverse().GetRotated(Vector(0.0f,0.0f,1.0f)).GetNormalized();
		const Vector vecLight = matViewingTransformation.GetInverse().GetRotated(Vector(0.0f,0.0f,1.0f)).GetNormalized();
		const Vector vecHalf = (vecLight + vecView).GetNormalized();
		const Matrix matInverse = matProjectionTransformation * (matViewingTransformation * matDataTransformation * hanVolume->GetTextureTransformation().GetInverse());

		m_shaVolume.SetOption("components",int(hanVolume->GetElementComponents()));
		m_shaVolume.SetOption("technique",std::string(GetPlugin().GetProperty("Technique")));
		m_shaVolume.SetOption("selectable",std::string(GetPlugin().GetProperty("Selectable")));
		m_shaVolume.bind();

		glUniform1iARB(m_shaVolume.GetUniformLocation("samFront"), 0);
		glUniform1iARB(m_shaVolume.GetUniformLocation("samBack"), 1);
		glUniform1iARB(m_shaVolume.GetUniformLocation("samVolume"), 2);
		glUniform1iARB(m_shaVolume.GetUniformLocation("samTransferFunction"), 3);
		glUniform1iARB(m_shaVolume.GetUniformLocation("samNoise"), 4);
		glUniform1iARB(m_shaVolume.GetUniformLocation("samAmbientTexture"), 5);
		glUniform1f(m_shaVolume.GetUniformLocation("fUnitStep"), 1.0f / hanVolume->GetTextureSize().GetMaximum());
		glUniform3fvARB(m_shaVolume.GetUniformLocation("vecSteps"), 1, hanVolume->GetTextureSize().GetInverse().Get());
		glUniform1f(m_shaVolume.GetUniformLocation("fSampleDistance"), fSampleDistance);
		glUniform3fvARB(m_shaVolume.GetUniformLocation("vecLight"), 1, vecLight.Get());
		glUniform3fvARB(m_shaVolume.GetUniformLocation("vecHalf"), 1, vecHalf.Get());
		glUniform1f(m_shaVolume.GetUniformLocation("z_cut"), z_cut);
		glUniform1f(m_shaVolume.GetUniformLocation("y_cut"), y_cut);
		glUniform1f(m_shaVolume.GetUniformLocation("x_cut"), x_cut);
		
		glUniform1i(m_shaVolume.GetUniformLocation("ray_steps"), ray_steps);
		glUniform1i(m_shaVolume.GetUniformLocation("n_rays"), n_rays);
		glUniform1i(m_shaVolume.GetUniformLocation("steplength"), steplength);



		glUniformMatrix4fv(m_shaVolume.GetUniformLocation("matVolume"),1,false,matVolumeTransformation.GetInverse().Get());
		glUniformMatrix4fv(m_shaVolume.GetUniformLocation("matInverse"),1,false,matInverse.Get());

		glBegin(GL_QUADS);
		glTexCoord2f(0.0f,0.0f);
		glVertex3f(-1.0f,-1.0f,0.0f);
		glTexCoord2f(1.0f,0.0f);
		glVertex3f(1.0f,-1.0f,0.0f);
		glTexCoord2f(1.0f,1.0f);
		glVertex3f(1.0f,1.0f,0.0f);
		glTexCoord2f(0.0f,1.0f);
		glVertex3f(-1.0f,1.0f,0.0f);
		glEnd();

		m_shaVolume.release();

		glActiveTexture(GL_TEXTURE5);		
		glBindTexture(GL_TEXTURE_3D,0);

		glActiveTexture(GL_TEXTURE4);
		glBindTexture(GL_TEXTURE_2D,0);

		glActiveTexture(GL_TEXTURE3);
		if (hanImage.IsValid()) hanImage->release();

		glActiveTexture(GL_TEXTURE2);
		hanVolume->release();

		glActiveTexture(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D,0);

		glActiveTexture(GL_TEXTURE0);		
		glBindTexture(GL_TEXTURE_2D,0);

		canCanvas.release();
	};

	void generateBounds()
	{
		if (m_bGenerateBounds)
		{
			Handle hanHandle = GetPlugin().GetProperty("Volume");

			if (!hanHandle.IsValid())
				return;
			
			Handle hanImage = GetPlugin().GetProperty("Transfer Function");
			Image<unsigned char,4> *pImage = dynamic_cast< Image<unsigned char,4> *>(hanImage.GetResource());

			if (!pImage)
			{
				static Image<unsigned char,4> imgImage;
				pImage = &imgImage;
			}

			std::vector<float> vecVertices;
			std::vector<unsigned int> vecIndices;

			RESOURCE_CALL(generateGeometry,*hanHandle,vecVertices,vecIndices,*pImage);

			if (vecVertices.size() > 0)
			{
				glBindBuffer(GL_ARRAY_BUFFER, m_uVertexBuffer);
				glBufferData(GL_ARRAY_BUFFER, vecVertices.size()*sizeof(float), (void*)&(vecVertices.front()), GL_STATIC_DRAW);
				glBindBuffer(GL_ARRAY_BUFFER, 0);
			}

			if (vecIndices.size() > 0)
			{
				glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_uIndexBuffer);
				glBufferData(GL_ELEMENT_ARRAY_BUFFER, vecIndices.size()*sizeof(unsigned int), (void*)&(vecIndices.front()), GL_STATIC_DRAW);
				glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
			}

			m_uElementCount = vecIndices.size();
			m_bGenerateBounds = false;
		}
	};

	void renderBounds()
	{
		generateBounds();

		glBindBuffer(GL_ARRAY_BUFFER, m_uVertexBuffer);
		glVertexPointer( 3, GL_FLOAT, 0, NULL);
		glEnableClientState(GL_VERTEX_ARRAY);	

		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_uIndexBuffer);
		glIndexPointer(GL_INT,0,NULL);
		glEnableClientState(GL_INDEX_ARRAY);

		glDrawElements(GL_QUADS,m_uElementCount,GL_UNSIGNED_INT,NULL);

		glDisableClientState( GL_VERTEX_ARRAY );	
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

		glDisableClientState(GL_INDEX_ARRAY);
		glBindBuffer(GL_ARRAY_BUFFER, 0);
	};

	virtual void propertyModifiedGenerate(const Variant & varVariant, const Observable::Event & eveEvent)
	{
		m_bGenerateBounds = true;
		GetPlugin().update();
	};

	virtual void propertyModifiedUpdate(const Variant & varVariant, const Observable::Event & eveEvent)
	{
		GetPlugin().update();
	};

	template < template<class,unsigned int> class ResourceType, class ElementType,unsigned int N>
	void generateGeometry(const ResourceType<ElementType,N> & resResource, std::vector<float> & vecVertices, std::vector<unsigned int> & vecIndices, const Image<unsigned char,4> & imgImage)
	{

	};

	template <class ElementType,unsigned int N>
	void generateGeometry(const Volume<ElementType,N> & volVolume, std::vector<float> & vecVertices, std::vector<unsigned int> & vecIndices, const Image<unsigned char,4> & imgImage)
	{
		Volume<ElementType,N>::HierachicalIterator iter(volVolume);
		static const unsigned int vuIndices[6][4] =
		{

			{0, 1, 2, 3},
			{3, 2, 6, 7},
			{7, 6, 5, 4},
			{4, 5, 1, 0},
			{5, 6, 2, 1},
			{7, 4, 0, 3}
		};

		const unsigned int uVolumeWidth = volVolume.GetWidth();
		const unsigned int uVolumeHeight = volVolume.GetHeight();
		const unsigned int uVolumeDepth = volVolume.GetDepth();

		const unsigned int uX = iter.GetPositionX();
		const unsigned int uY = iter.GetPositionY();
		const unsigned int uZ = iter.GetPositionZ();

		const unsigned int uWidth = iter.GetWidth();
		const unsigned int uHeight = iter.GetHeight();
		const unsigned int uDepth = iter.GetDepth();

		const unsigned int uStartX = uX - uWidth/2;
		const unsigned int uStartY = uY - uHeight/2;
		const unsigned int uStartZ = uZ - uDepth/2;

		if (uStartX > volVolume.GetWidth() || uStartY > volVolume.GetHeight() || uStartZ > volVolume.GetDepth())
			return;

		const unsigned int uEndX = std::min(uVolumeWidth,uX + uWidth/2);
		const unsigned int uEndY = std::min(uVolumeHeight,uY + uHeight/2);
		const unsigned int uEndZ = std::min(uVolumeDepth,uZ + uDepth/2);

		float vfVertices[8][3];
		vfVertices[0][0] = vfVertices[1][0] = vfVertices[2][0] = vfVertices[3][0] = float(uStartX);
		vfVertices[4][0] = vfVertices[5][0] = vfVertices[6][0] = vfVertices[7][0] = float(uEndX) - 1.0f/16384.0f;
		vfVertices[0][1] = vfVertices[1][1] = vfVertices[4][1] = vfVertices[5][1] = float(uStartY);
		vfVertices[2][1] = vfVertices[3][1] = vfVertices[6][1] = vfVertices[7][1] = float(uEndY) - 1.0f/16384.0f;
		vfVertices[0][2] = vfVertices[3][2] = vfVertices[4][2] = vfVertices[7][2] = float(uStartZ);
		vfVertices[1][2] = vfVertices[2][2] = vfVertices[5][2] = vfVertices[6][2] = float(uEndZ) - 1.0f/16384.0f;


		const unsigned int uOffset = vecVertices.size() / 3;

		for (int i=0;i<24;i++)
		{
			vecVertices.push_back(*(&(vfVertices[0][0])+i));
			vecIndices.push_back(*(&(vuIndices[0][0])+i)+uOffset);
		}
	};

	template <class ElementType>
	void generateGeometry(const Volume<ElementType,1> & volVolume, std::vector<float> & vecVertices, std::vector<unsigned int> & vecIndices, const Image<unsigned char,4> & imgImage)
	{
		Volume<ElementType,1>::HierachicalIterator iter(volVolume);
		Image<unsigned char,4>::IntegralIterator intIntegral(imgImage);
		generateGeometry(volVolume,vecVertices,vecIndices,imgImage,iter);
	};

	template <class ElementType>
	void generateGeometry(const Volume<ElementType,1> & volVolume, std::vector<float> & vecVertices, std::vector<unsigned int> & vecIndices, const Image<unsigned char,4>::IntegralIterator & intIntegral, typename const Volume<ElementType,1>::HierachicalIterator & iter)
	{
		const Voxel<float,1> voxMinimum = iter.GetMinimum();
		const Voxel<float,1> voxMaximum = iter.GetMaximum();

		const unsigned int u0x = unsigned int(voxMinimum.Get(0)*float(intIntegral.GetWidth()-1));
		const unsigned int u1x = unsigned int(voxMaximum.Get(0)*float(intIntegral.GetWidth()-1));
		const Voxel<float,2> voxIntegral = intIntegral.GetIntegral(u0x,u1x);

		if (voxIntegral.Get(3) > 1.192092896e-07F)
		{
			if (iter.IsLeaf())
			{
				static const unsigned int vuIndices[6][4] =
				{

					{0, 1, 2, 3},
					{3, 2, 6, 7},
					{7, 6, 5, 4},
					{4, 5, 1, 0},
					{5, 6, 2, 1},
					{7, 4, 0, 3}
				};

				const unsigned int uVolumeWidth = volVolume.GetWidth();
				const unsigned int uVolumeHeight = volVolume.GetHeight();
				const unsigned int uVolumeDepth = volVolume.GetDepth();

				const unsigned int uX = iter.GetPositionX();
				const unsigned int uY = iter.GetPositionY();
				const unsigned int uZ = iter.GetPositionZ();

				const unsigned int uWidth = iter.GetWidth();
				const unsigned int uHeight = iter.GetHeight();
				const unsigned int uDepth = iter.GetDepth();

				const unsigned int uStartX = uX - uWidth/2;
				const unsigned int uStartY = uY - uHeight/2;
				const unsigned int uStartZ = uZ - uDepth/2;

				if (uStartX > volVolume.GetWidth() || uStartY > volVolume.GetHeight() || uStartZ > volVolume.GetDepth())
					return;

				const unsigned int uEndX = std::min(uVolumeWidth,uX + uWidth/2);
				const unsigned int uEndY = std::min(uVolumeHeight,uY + uHeight/2);
				const unsigned int uEndZ = std::min(uVolumeDepth,uZ + uDepth/2);

				float vfVertices[8][3];
				vfVertices[0][0] = vfVertices[1][0] = vfVertices[2][0] = vfVertices[3][0] = float(uStartX);
				vfVertices[4][0] = vfVertices[5][0] = vfVertices[6][0] = vfVertices[7][0] = float(uEndX) - 1.0f/16384.0f;
				vfVertices[0][1] = vfVertices[1][1] = vfVertices[4][1] = vfVertices[5][1] = float(uStartY);
				vfVertices[2][1] = vfVertices[3][1] = vfVertices[6][1] = vfVertices[7][1] = float(uEndY) - 1.0f/16384.0f;
				vfVertices[0][2] = vfVertices[3][2] = vfVertices[4][2] = vfVertices[7][2] = float(uStartZ);
				vfVertices[1][2] = vfVertices[2][2] = vfVertices[5][2] = vfVertices[6][2] = float(uEndZ) - 1.0f/16384.0f;


				const unsigned int uOffset = vecVertices.size() / 3;

				for (int i=0;i<24;i++)
				{
					vecVertices.push_back(*(&(vfVertices[0][0])+i));
					vecIndices.push_back(*(&(vuIndices[0][0])+i)+uOffset);
				}
			}
			else
			{
				for (Volume<ElementType,1>::HierachicalIterator i(iter); !i.IsAtEnd(); ++i)
					generateGeometry(volVolume,vecVertices,vecIndices,intIntegral,(*i));
			}
		}
	};

private:

	int ray_steps;
	int n_rays;
	int steplength;

	float z_scale;
	float z_cut;
	float y_cut;
	float x_cut;

	unsigned int m_uWidth;
	unsigned int m_uHeight;

	unsigned int m_uVolumeFrameBuffer;
	unsigned int m_uFrontColorTexture;
	unsigned int m_uBackColorTexture;
	unsigned int m_uVolumeColorTexture;
	unsigned int m_uGeometryDepthTexture;
	unsigned int m_uVolumeDepthRenderBuffer;
	unsigned int m_uNoiseTexture;
	unsigned int m_uAmbientTexture;

	unsigned int m_uVertexBuffer;
	unsigned int m_uIndexBuffer;
	unsigned int m_uElementCount;
	bool m_bGenerateBounds;

	Shader m_shaVolume;
	Shader m_shaGeometry;
	Shader m_shaEntry;

	ModifiedObserver m_modVariantObserverGenerate;
	ModifiedObserver m_modVariantObserverUpdate;
};

DECLARE_RENDERERPLUGIN(RendererVolumeLao)